#include <stdio.h>
#include <avr/io.h>
#include <util/delay.h>
#include <inttypes.h>
#include <string.h>

#include "game.h"
#include "vector.h"
#include "ST7735R_TFT.h"
#include "ST7735R_TFT.c"
#include "sound.c"

#define SCORE_OFFSET       20

enum food_type_e
{
	BLUE,
	RED
};

enum game_status_e
{
	GAME_INIT,
	GAME_RUNNING,
	GAME_LOST,
	GAME_WON,
};

enum food_status_e
{
	FOOD_EATEN,
	FOOD_FRESH
};

enum snake_direction_e
{
	UP,
	DOWN,
	LEFT,
	RIGHT
};

typedef struct color_t
{
	uint8_t r;
	uint8_t g;
	uint8_t b;
} color_t;

typedef struct food_t
{

	color_t color;
	vector2i_t pos;
	uint8_t type;
	uint8_t status;
} food_t;

typedef struct segment_t
{
	color_t color;
	vector2i_t pos;
	uint8_t    head;
} segment_t;

typedef struct snake_t
{
	
	segment_t   segments[40];
	segment_t   tail;
	uint8_t     length;
	uint8_t     direction;
} snake_t;

typedef struct game_t
{

	uint8_t     status;
	snake_t     snake;
	food_t food;
	int score;
} game_t;

game_t game;
int current_food_items = 0;


void draw_segment(int x, int y, uint8_t r, uint8_t g, uint8_t b);
void generate_food();
void draw_food();
void physics(void);
int  check_food_collision(void);

/****************************************************************************
* Public Functions
****************************************************************************/

void init(void)
{
	ST7735R_FillRect(0, 0, ST7735R_WIDTH - 1, ST7735R_HEIGHT - 1, 0, 0, 0);

	/* Initial snake length and direction. */
	game.snake.direction = RIGHT;
	game.snake.length = 6;
	game.status = GAME_INIT;
	game.score = 0;

	int i = 0;
	while(i < 6) {

		if (i == 0) {
			game.snake.segments[i].color.r = 0;
			game.snake.segments[i].color.g = 255;
			game.snake.segments[i].color.b = 0;

			game.snake.segments[i].head = 1;
			game.snake.segments[i].pos.x = 60;
			game.snake.segments[i].pos.y = 75;

		} else {
			game.snake.segments[i].color.r = 0;
			game.snake.segments[i].color.g = 255;
			game.snake.segments[i].color.b = 0;

			game.snake.segments[i].head = 0;
			game.snake.segments[i].pos.x = 60 - i * 6 - i * 2;
			game.snake.segments[i].pos.y = 75;
		}

		i++;
	}
	
	/* Generate food on the map. */
	generate_food();
	
	/* Draw a line to split the score and game area */
	ST7735R_FillRect(0, 0, ST7735R_WIDTH - 1, ST7735R_HEIGHT - 1, 0, 0, 0);
	ST7735R_Line(0, SCORE_OFFSET - 2, ST7735R_WIDTH,  SCORE_OFFSET - 2, 0, 0, 0);


	draw_segment(game.snake.tail.pos.x, game.snake.tail.pos.y,
	0, 0, 0);

	i = 0;
	while (i < game.snake.length - 1 && i < 2) {
		if (game.status == GAME_RUNNING) {
			break;
		}
		
		int pos_x = game.snake.segments[i].pos.x;
		int pos_y = game.snake.segments[i].pos.y;
		uint8_t r = game.snake.segments[i].color.r;
		uint8_t g = game.snake.segments[i].color.g;
		uint8_t b = game.snake.segments[i].color.b;
		
		draw_segment(pos_x, pos_y, r, g, b);

		if (!i)
			ST7735R_FilledCircle(pos_x + 3, pos_y + 3 + SCORE_OFFSET, 1,
			0,
			0,
			0);
	i++;
	}
	draw_food();

	/* Start the game! */
	game.status = GAME_RUNNING;
}

void move_snake(int x, int y)
{
	
	if (abs(x) > abs(y))
	{
		if (x < 0) {
			if (game.snake.direction != LEFT)
			game.snake.direction = RIGHT;
			} else if(game.snake.direction != RIGHT)
				game.snake.direction = LEFT;

		} else if (abs(x) < abs(y)) {
		if (y < 0) {
			if (game.snake.direction != UP)
			game.snake.direction = UP;
			} else 
			game.snake.direction = DOWN;
		
	} else
		return;
	
}


void loop(void)
{
	/* Check if the player lost or won. */
	int i = 0;
	char buf[30];

	switch(game.status) {
		case  GAME_LOST:	
						ST7735R_DrawText(ST7735R_WIDTH / 2 - 55, ST7735R_HEIGHT / 2 - 20,
						"NU TE MANCA SINGUR!", 255, 255, 255, 0, 0, 0);
						make_sound(1000, 0);
						_delay_ms(3000);
						init();
						break;

		case GAME_WON:
						ST7735R_DrawText(ST7735R_WIDTH / 2 - 40, ST7735R_HEIGHT / 2 - 20,
						"Bine bossanova!", 255, 255, 255, 0, 0, 0);
						make_sound(0, 1);
						_delay_ms(4000);
						init();

		default: break;
	}

	physics();
	draw_segment(game.snake.tail.pos.x, game.snake.tail.pos.y,
	0, 0, 0);

	/* Draw first 2 segments. First is new head's location,
	second to cover old head's position. */
	while(i < game.snake.length - 1) {
		if (game.status == GAME_RUNNING && i >= 2) {
			break;
		}
		
		int pos_x = game.snake.segments[i].pos.x;
		int pos_y = game.snake.segments[i].pos.y;
		uint8_t r = game.snake.segments[i].color.r;
		uint8_t g = game.snake.segments[i].color.g;
		uint8_t b = game.snake.segments[i].color.b;
		
		draw_segment(pos_x, pos_y, r, g, b);

		if (i == 0)
			ST7735R_FilledCircle(pos_x + 3, pos_y + 3 + SCORE_OFFSET, 1,
			0,
			0,
			0);
		i++;
	}
	sprintf(buf, "SCORE: %d", game.score);
	ST7735R_DrawText(5, 6, buf, 255, 255, 255, 0, 0, 0);
}


void draw_segment(int x, int y, uint8_t r, uint8_t g, uint8_t b)
{
	uint8_t segment_x = 0;
	uint8_t segment_y = 0;
	
	if (x + 6 > 120) {
		segment_x = 1;
	}
	
	if (y + 6 > 150 - 1) {
		segment_y = 1;
	}
	
	if (segment_x == 0 && segment_y == 0) {
		ST7735R_FillRect(x, y + SCORE_OFFSET, x + 6, y + 6 + SCORE_OFFSET, r, g, b);
	} else if (segment_y == 0) {
		ST7735R_FillRect(x, y + SCORE_OFFSET, 120, y + 6 + SCORE_OFFSET, r, g, b);
	} else
		ST7735R_FillRect(x, y + SCORE_OFFSET, x + 6, y + 6 + SCORE_OFFSET, r, g, b);

}

void physics(void)
{
	int len;
	len = game.snake.length - 1;
	int buzz_delay;
	buzz_delay = 0;
	int food_collision_index;
	food_collision_index = -1;
	int i = 0;

	if ((food_collision_index = check_food_collision()) >= 0)
	{
		game.food.status = FOOD_EATEN;
		
		if (game.food.type == BLUE) {
			game.score += 10;
		} else {
			game.score += 5;
		}
		draw_food();
		generate_food();
		draw_food();
		draw_segment(game.snake.tail.pos.x, game.snake.tail.pos.y,
		0, 0, 0);

		while(i < game.snake.length - 1) {
			if (game.status == GAME_RUNNING && i >= 2) {
				break;
			}
		
			int pos_x;
			int pos_y;
			uint8_t r;
			uint8_t g;
			uint8_t b;
		
			pos_x = game.snake.segments[i].pos.x;
			pos_y = game.snake.segments[i].pos.y;
			r = game.snake.segments[i].color.r;
			g = game.snake.segments[i].color.g;
			b = game.snake.segments[i].color.b;

			draw_segment(pos_x, pos_y, r, g, b);

			if (i == 0)
				ST7735R_FilledCircle(pos_x + 3, pos_y + 3 + SCORE_OFFSET, 1,
				0,
				0,
				0);
			i++;
		}
			game.snake.segments[len+1].color.r = 0;
			game.snake.segments[len+1].color.g = 255;
			game.snake.segments[len+1].color.b = 0;

			game.snake.segments[len+1].head = 0;
			game.snake.segments[len+1].pos.x = game.snake.segments[len].pos.x;
			game.snake.segments[len+1].pos.y = game.snake.segments[len].pos.y;
		game.snake.length++;
		buzz_delay = 100;
	}
	
	segment_t head = game.snake.segments[0];
	i = 1;
	while(i < game.snake.length - 1) {

		segment_t snake = game.snake.segments[i];
		
		if ((snake.pos.x + 6) >= head.pos.x
		&& snake.pos.x <= (head.pos.x + 6)
		&& (snake.pos.y + 6) >= head.pos.y
		&& snake.pos.y <= (head.pos.y + 6))
		game.status = GAME_LOST;
		i++;
	}

	if (game.snake.length == 50)
		game.status = GAME_WON;
	else {
		len = game.snake.length - 1;
		game.snake.tail.pos.x = game.snake.segments[len].pos.x;
		game.snake.tail.pos.y = game.snake.segments[len].pos.y;
		i = len;
		while (i > 0) {
			game.snake.segments[i].pos.x = game.snake.segments[i - 1].pos.x;
			game.snake.segments[i].pos.y = game.snake.segments[i - 1].pos.y;
			i--;
		}
		
		int pos_x = game.snake.segments[0].pos.x;
		int pos_y = game.snake.segments[0].pos.y;
		uint8_t direction = game.snake.direction;
		
		switch(direction) {
			case RIGHT:game.snake.segments[0].pos.x = ( pos_x + 6 + 2 + 120)
			% 120;
			break;

			case LEFT:
			game.snake.segments[0].pos.x = ( pos_x - 6 - 2 + 120)
			% 120;
			break;

			case UP:
			game.snake.segments[0].pos.y = ( pos_y + 6 + 2 + 150)
			% 150;
			break;

			case DOWN:
			game.snake.segments[0].pos.y = ( pos_y - 6 - 2 + 150)
			% 150;
			break;

			default: break;
		}
	}
	if (buzz_delay != 0)
		make_sound(100, 0);
}


void generate_food()
{
	int i;
	int pos_x;
	int pos_y;
	int ok = 0;
	int once = 0;

	while (ok == 0) {
		ok = 1;
		pos_x = abs(rand()) % 120;
		pos_y = abs(rand()) % 130;
		int food_x = pos_x - 4;
		int food_y = pos_y - 4;
		int width = 2 * 4;
		
		if (once == 0) {
			segment_t snake = game.snake.tail;
		
			if ((snake.pos.x + 6) >= food_x
			&& snake.pos.x <= (food_x + width)
			&& (snake.pos.y + 6) >= food_y
			&& snake.pos.y <= (food_y + width))
				ok = 0;
			once++;
		}
		
		for (i = 0; i < game.snake.length - 1; i++) {

			segment_t snake = game.snake.segments[i];
			
			if ((snake.pos.x + 6) >= food_x
			&& snake.pos.x <= (food_x + width)
			&& (snake.pos.y + 6) >= food_y
			&& snake.pos.y <= (food_y + width))
			ok = 0;
		}
		
		
			food_t *other_food = &game.food;
			int other_food_x = other_food->pos.x - 4;
			int other_food_y = other_food->pos.y - 4;
			
			if ((other_food_x + width) >= food_x
			&& other_food_x <= (food_x + width)
			&& (other_food_y + width) >= food_y
			&& other_food_y <= (food_y + width))
			ok = 0;
		
	}
	
	current_food_items++;
	food_t *food = &game.food;
		
		food->type = RED;
		food->color.r = 255;
		food->color.g = 0;
		food->color.b = 0;

	
	food->pos.x = pos_x;
	food->pos.y = pos_y;
	
	food->status = FOOD_FRESH;
}


void draw_food()
{
	food_t *food = &game.food;
	if (food->status == FOOD_EATEN)
		ST7735R_FilledCircle(food->pos.x, food->pos.y + SCORE_OFFSET, 4, 0, 0, 0);
	else
		ST7735R_FilledCircle(food->pos.x, food->pos.y + SCORE_OFFSET, 4,
		food->color.r,
		food->color.g,
		food->color.b);
}

int check_food_collision(void)
{
	segment_t head = game.snake.segments[0];
	
	/* Consider our food circle a box with width 2 * radius */

	
		food_t *food = &game.food;
		int food_x = food->pos.x - 4;
		int food_y = food->pos.y - 4;
		int width = 2 * 4;
		
		if ((head.pos.x + 6) >= food_x
		&& head.pos.x <= (food_x + width)
		&& (head.pos.y + 6) >= food_y
		&& head.pos.y <= (food_y + width))
		return 1;
	
	return -1;
}



